package RPM::SpecEditor;
our $VERSION = 0.002;

use strict;
use warnings;
use RPM::SpecEditor::Node;

=encoding utf8

=head1 NAME

RPM::SpecEditor - RPM specification parser and editor

=head1 DESCRIPTION

This module allows to parse and edit RPM specification files.

Please be ware that it's impossible to interpret RPM macros without their definitions. Therefore the specification handling can be inaccurate.

=head2 new(handle => IO_HANDLE)

=head2 new(file => FILE_NAME)

=head2 new(string => STRING_VALUE)

Parse a specification file from IO::Handle object or file or a string. The
input must be in UTF-8. It returns an object that can be inspected by other
methods. In case of error it dies.

=cut

sub new {
    my ($class, %opts) = @_;

    my $self = bless {}, $class;

    my $handle;
    if (exists $opts{file}) {
        $self->_open_file($opts{file}, '<');
    } elsif (exists $opts{string}) {
        require IO::String;
        $self->{private_handle} = 1;
        $self->{handle} = IO::String->new($opts{string});
    } elsif (exists $opts{handle}) {
        $self->{private_handle} = 0;
        $self->{handle} = $opts{handle};
    } else {
        die "Missing handle, file or string argument";
    }

    $self->_parse();

    if ($self->{private_handle}) {
        $self->{handle}->close;
        delete $self->{handle};
    }

    return $self;
}


# Open a file from a file name. Dies on error.
sub _open_file {
    my ($self, $name, $mode) = @_;
    my $handle;
    open($handle, $mode, $name)
        or die "Could not open $name: $!";
    $self->{private_handle} = 1;
    $self->{handle} = $handle;
    $self->{file_name} = $name;
    return $self;
}

=head2 dom()

Return document object module of the specification file. It's a tree of
L<RPM::SpecEditor::Node> object. The tree itself if L<RPM::SpecEditor::Node>
object.

=cut

sub dom {
    return shift->{dom};
}

# Build DOM tree from a handle.
sub _parse {
    my ($self) = @_;

    binmode $self->{handle}, ':encoding(UTF-8)';
    my @lines = readline $self->{handle};
    if ($!) {
        if (exists $self->{file_name}) {
            die "Error while reading from file " . $self->{file_name} . ": $!";
        } else {
            die "Error while reading from a file handle: $!";
        }
    }

    for (@lines) { chomp };
    $self->{lines} = \@lines;
    $self->{dom} = RPM::SpecEditor::Node->new(type => 'root');

    for (my $index = 0; $index < scalar @{$self->{lines}}; $index++) {
        $self->_parse_line($index);
    }

    $self->_parse_tree();

    return $self;
}

# Parses lines into syntax tree
sub _parse_line {
    my ($self, $line_number) = @_;

    my %node = (
        line => $line_number,
    );
    if ($self->{lines}[$line_number] =~ /^(\s*)\#(\s*)(.*)/) {
        $node{type} = 'comment';
        $node{leftpad} = $1;
        $node{separator} = $2;
        $node{value} = $3;
    } elsif ($self->{lines}[$line_number] =~ /^(\s*)(\w+):(\s*)(.*+)\s*$/) {
        $node{type} = 'tag';
        $node{leftpad} = $1;
        $node{name} = $2;
        $node{separator} = $3;
        $node{value} = $4;
    } else {
        $node{type} = 'line';
        $node{value} = $self->{lines}[$line_number];
    }

    $self->{dom}->addChild(RPM::SpecEditor::Node->new(%node));

    return $self;
}

# Parses syntax tree into semantic tree.
sub _parse_tree {
    my $self = shift;

    my $node = $self->dom->getChild(0);
    if (!defined $node) { return $self; }

    do {
        $self->_expand_dependency($node);
    } while (defined($node = $node->nextSibling));
}

# If it's a tag node for a dependency, parses it's value and replaces it with
# a dependency subtree
sub _expand_dependency {
    my ($self, $node) = @_;

    # Check it's a dependency tag
    if ($node->type ne 'tag') {
        return $self;
    }
    if (!$node->isDependencyTag) {
       return $self;
    }

    # Prepare new subtree
    my $dependency_root = RPM::SpecEditor::Node->new(
        line => $node->line - 1,
        type => $node->name,
        leftpad => $node->leftpad,
        separator => $node->separator,
        value => $node->value
    );

    my @tokens = _tokenize_dependency($node->value);
    if (!@tokens) {
        die "Missing a dependency after " . $dependency_root->type .
            "on line " . $dependency_root->line . ": " .
            $self->{lines}[$dependency_root->line - 1];
    }
    # Parse tokens to list of:
    # symbol [[operatorleftpad operator versionleftpad version] rightpad]
    my %dependency;
    while (defined(my $token = shift @tokens)) {
        %dependency = ();
        $dependency{line} = $dependency_root->line - 1;
        $dependency{type} = 'dependency';

        # Assume the first token is always a word
        # Dependency symbol
        if (!defined $token) {
            die "Missing a dependency after " . $dependency_root->type .
                "on line " . $dependency_root->line . ": " .
                $self->{lines}[$dependency_root->line - 1];
        }
        $dependency{symbol} = $token;
        $token = shift @tokens;

        # Dependency-operator separator
        if (defined $token) {
            $dependency{operatorleftpad} = $token;
            $token = shift @tokens;

            if (!defined $token) {
                $dependency{rightpad} = delete $dependency{operatorleftpad};
                # "continue" block is not executed after "last" command.
                my $dependency = RPM::SpecEditor::Node->new(
                    line => $dependency_root->line - 1,
                    type => 'dependency',
                    %dependency
                );
                $dependency_root->addChild($dependency);
                last;
            }

            # Next dependency
            if ($token !~ /^[=><]/) {
                $dependency{rightpad} = delete $dependency{operatorleftpad};
                unshift @tokens, $token;
                next;
            }

            # Operator
            $dependency{operator} = $token;
            $token = shift @tokens;

            if (!defined $token) {
                die "Missing a space and a version after " .
                    $dependency{operator} .
                    "on line " . $dependency_root->line . ": " .
                    $self->{lines}[$dependency_root->line - 1];
            }
            $dependency{versionleftpad} = $token;
            $token = shift @tokens;

            # Version
            if (!defined $token) {
                die "Missing a version after dependency operator " .
                    $dependency{operator} .
                    "on line " . $dependency_root->line . ": " .
                    $self->{lines}[$dependency_root->line - 1];
            }
            $dependency{version} = $token;
            $token = shift @tokens;

            # Rightpad
            if (defined $token) {
                $dependency{rightpad} = $token;
            }
        }
    } continue {
        my $dependency = RPM::SpecEditor::Node->new(
            line => $dependency_root->line - 1,
            type => 'dependency',
            %dependency
        );
        $dependency_root->addChild($dependency);
    }

    # Replace the tag node with a dependency subtree
    $node->replaceContent($dependency_root);

    return $self;
}

# Return list of word-or-separator tokens parsed from a string.
sub _tokenize_dependency {
    my $input = shift;

    my @tokens;
    my $in_word = 1;
    for my $character (split '', $input) {
        my $append = 1;
        if ($in_word) {
            if ($character =~ /[\s,]/) {
                $append = 0;
            }
        } elsif ($character !~ /[\s,]/) {
            $append = 0;
        }
        if ($append) {
            if (@tokens) {
                $tokens[-1] .= $character;
            } else {
                @tokens = ($character);
            }
        } else {
            $in_word = !$in_word;
            push @tokens, $character;
        }
    }
    return @tokens;
}


=head2 save(handle => IO_HANDLE)

=head2 save(file => FILE_NAME)

=head2 save(string => STRING_REFERENCE)

Serialize parsed specification tree into a L<IO::Handle>, a file or a string
scalar.

The output medium will be emptied and and the output data will be written in
UTF-8. Be ware that output to a string will save the data as a Unicode string.
Not as a byte stream.

=cut

sub save {
    my ($self, %opts) = @_;

    my $handle;
    if (exists $opts{file}) {
        $self->_open_file($opts{file}, '>');
    } elsif (exists $opts{string}) {
        ${$opts{string}} = '';
        require IO::String;
        $self->{private_handle} = 1;
        $self->{handle} = IO::String->new($opts{string});
    } elsif (exists $opts{handle}) {
        eval { $opts{handle}->truncate(0) };
        $self->{private_handle} = 0;
        $self->{handle} = $opts{handle};
    } else {
        die "Missing handle, file or string argument";
    }

    $self->_serialize();

    if ($self->{private_handle}) {
        $self->{handle}->close;
        delete $self->{handle};
    }

    return $self;
}


# Serialize DOM tree to a handle.
sub _serialize {
    my ($self) = @_;

    binmode $self->{handle}, ':encoding(UTF-8)';

    for my $line ($self->_serialize_tree()) {
        undef $!;
        $self->{handle}->print($line . "\n");
        if ($!) {
            if (exists $self->{file_name}) {
                die "Error while writing to file " . $self->{file_name} .
                    ": $!";
            } else {
                die "Error while writing to a file handle: $!";
            }
        }
    }

    return $self;
}


# Serialize semantic tree into list of lines without end-of-line trailers.
sub _serialize_tree {
    my $self = shift;

    my $node = $self->dom->getChild(0);
    if (!defined $node) { return; }

    my @lines;
    do {
        push @lines, $node->serialize();
    } while (defined($node = $node->nextSibling));

    return @lines;
}

1;

=head1 AUTHOR

Petr Písař <ppisar@redhat.com>

=head1 COPYING

Copyright (C) 2016, 2017  Petr Písař <ppisar@redhat.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
