package RPM::SpecEditor::Node;
our $VERSION = 0.002;

use strict;
use warnings;
use parent 'Tree::Simple';

=encoding utf8

=head1 NAME

RPM::SpecEditor::Node - RPM specification semantical node

=head1 DESCRIPTION

L<RPM::SpecEditor::Node> represents a node in document object model tree
parsed from a specification file.

You can call any L<Tree::Simple> methods on the node. E.g to append the node to a paretn node or to traverse the DOM tree.

=head2 new(HASH_OF_ATTRIBUTES)

Create a new node of specification DOM. Attributes are stored inside the node.

=cut

sub new {
    my ($class, %spec) = @_;

    my $self = $class->SUPER::new(\%spec);

    return $self;
}

=head2 nextSibling()

Return a sibbling right of the current node. Return undef if there is none.

=cut

sub nextSibling {
    my $self = shift;
    return $self->getSibling($self->getIndex + 1);
}

=head2 previousSibling()

Return a sibbling left of the current node. Return undef if there is none.

=cut

sub previousSibling {
    my $self = shift;
    my $index = $self->getIndex;
    if ($index < 1) {
        return undef;
    }
    return $self->getSibling($index - 1);
}

=head2 replaceNode(NEW_NODE)

Replace the node with given L<RPM::SpecEditor::Node> new node and its children.
The original node will be disconnected from its tree.

=cut

sub replaceNode {
    my ($self, $new) = @_;

    my $index = $self->getIndex;
    my $parent = $self->getParent;
    $parent->removeChild($index);
    $parent->insertChild($index, $new);

    return $self;
}

=head2 replaceContent(NEW_NODE)

Replace the node's attributes and children  with given L<RPM::SpecEditor::Node> new node's attributes and children. The node identity will be preserved.

=cut

sub replaceContent {
    my ($self, $new) = @_;

    # Replace value
    $self->setNodeValue($new->getNodeValue);
    
    # Replace children
    my $old_count = $self->getChildCount;
    for (1 .. $old_count) { $self->removeChild(0) };
    for my $child ($new->getAllChildren) {
        $new->removeChild(0);
        $self->addChildren($child);
    }

    return $self;
}

=head2 line()

Ruturn line number of the node in the specification file. The lines are
numbered from 1.

=cut

sub line {
    return shift->getNodeValue()->{line} + 1;
}

=head2 type()

Return node type. Known values are:

=over 4

=item line

=item comment

=item tag

=item Provides

=item BuildRequires

=item BuildConflicts

=item Requires

=item Recommends

=item Suggests

=item Supplements

=item Conflicts

=item Obsoletes

=item dependency

=back

=cut

sub type {
    return shift->getNodeValue()->{type};
}

=head2 name()

=head2 name(NEW_VALUE)

Get ot set a name of the node. This applies to I<tag> type and it carry value
like C<Requires> or C<BuildRequires>.

=cut

sub name {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{name} = shift;
    }
    return $self->getNodeValue()->{name};
}

=head2 value()

=head2 value(NEW_VALUE)

Get or set a string value of the node.

=cut

sub value {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{value} = shift;
    }
    return $self->getNodeValue()->{value};
}

=head2 leftpad()

=head2 leftpad(NEW_VALUE)

Get or set a string value of left white-space padding of the node. E.g. this
is the leading space before C<#> character in a comment. Or the leading space
before a tage name.

=cut

sub leftpad {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{leftpad} = shift;
    }
    return $self->getNodeValue()->{leftpad};
}

=head2 separator()

=head2 separator(NEW_VALUE)

Get or set a string value of a white-space separator between the name/marker
and the value of the node. E.g. this is the space between C<#> and C<A>
character in the C<# A> comment. Or the space between tage name and and tag
value.

=cut

sub separator {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{separator} = shift;
    }
    return $self->getNodeValue()->{separator};
}

=head2 symbol()

=head2 symbol(NEW_VALUE)

Get or set a dependency RPM symbol value. E.g. I<<Requires: A  >= 1>> has I<A>
symbol.

=cut

sub symbol {
    my $node = shift;
    if (@_) {
        $node->getNodeValue()->{symbol} = shift;
    }
    return $node->getNodeValue()->{symbol};
}

=head2 operatorLeftPad()

=head2 operatorLeftPad(NEW_VALUE)

Get or set a separator between a dependency symbol and its operator.
E.g. I<<Requires: A  >= 1>> has two spaces.

=cut

sub operatorLeftPad {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{operatorleftpad} = shift;
    }
    return $self->getNodeValue()->{operatorleftpad};
}

=head2 operator()

=head2 operator(NEW_VALUE)

Get or set an operator between a dependency symbol its version.
E.g. I<<Requires: A  >= 1>> has I<<>=>>.

=cut

sub operator {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{operator} = shift;
        if (defined $self->getNodeValue()->{operator}) {
            if (!defined $self->operatorLeftPad) {
                $self->operatorLeftPad(' ');
            }
        }
    }
    return $self->getNodeValue()->{operator};
}

=head2 versionLeftPad()

=head2 versionLeftPad(NEW_VALUE)

Get or set a separator between a dependency operator and its version.
E.g. I<<Requires: A  >= 1>> has one space.

=cut

sub versionLeftPad {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{versionleftpad} = shift;
    }
    return $self->getNodeValue()->{versionleftpad};
}

=head2 version()

=head2 version(NEW_VALUE)

=head2 version(undef)

Get or set or remove a dependency version. Removing version will also remove
operators and their paddings. E.g. I<<Requires: A  >= 1>> has I<1>.

=cut

sub version {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{version} = shift;
        if (!defined $self->getNodeValue()->{version}) {
            $self->getNodeValue()->{versionleftpad} = undef;
            $self->getNodeValue()->{operator} = undef;
            $self->getNodeValue()->{operatorleftpad} = undef;
        } else {
            if (!defined $self->versionLeftPad) { $self->versionLeftPad(' '); }
            if (!defined $self->operator) { $self->operator('>='); }
        }
    }
    return $self->getNodeValue()->{version};
}

=head2 rightPad()

=head2 rightPad(NEW_VALUE)

Get or set a seperator between this and following dependency symbol. Or
a traling space after the last dependency symbol. E.g. I<Requires: A, B> has
a comma filloed by a space.

=cut

sub rightPad {
    my $self = shift;
    if (@_) {
        $self->getNodeValue()->{rightpad} = shift;
    }
    return $self->getNodeValue()->{rightpad};
}

=head2 copyDefaultSpacing(TEMPLATE_NODE)

If the node has some undefined space attributes (paddings, separators), it will copy them from a template node. If the template is missing them or the template is undef, sets some default values.

Only tag and and dependency tags are supported now.

=cut 

sub copyDefaultSpacing {
    my ($self, $template) = @_;

    # Provide default leftpad and separator
    if (!defined $self->leftpad) {
        if (defined $template) {
            $self->leftpad($template->leftpad);
        } else {
            $self->leftpad('');
        }
    }
    if (!defined $self->separator) {
        if (defined $template) {
            $self->separator($template->separator);
        } else {
            $self->separator(' ');
        }
    }

    return $self;
}

=head2 remove()

Remove the node. It can remove other nodes if syntax require is. E.g. removing
last dependency will remove its parent too.

=cut

sub remove {
    my $self = shift;

    my $parent = $self->getParent;
    my $remove_parent = 0;

    # Handle dependency list
    if ($self->type eq 'dependency') {
        # Remove rightPad from left sibling if we are the last sibling.
        if (!$self->nextSibling) {
            my $previous_dependency = $self->previousSibling;
            if ($previous_dependency) {
                $previous_dependency->rightPad(undef);
            } else {
                $remove_parent = 1;
            }
        }
    }

    # Remove the node
    $parent->removeChild($self->getIndex);

    # Remove parent if that was the last dependency child
    if ($remove_parent) {
        $parent->remove;
    }

    return $self;
}

=head2 insertBefore(NEW_NODE)

Insert a new L<RPM::SpecEditor::Node> node before this node. If this is
a dependency node, then the new node must also be a dependency. Unspecified
pads will be copied from a sibling or a space and comma-space will be used.

=cut

sub insertBefore {
    my ($self, $new) = @_;

    if ($new->type eq 'dependency') {
        if ($self->type ne 'dependency') {
            die "Inserting a dependency node next to " . $new->type .
                " node violates syntax. Insert a tag or a Requires-like " .
                "node first.";
        }
        # Provide default rightPad
        if (!defined $new->rightPad) {
            $new->rightPad($self->rightPad // ', ');
        }
        if (defined $new->version) {
            # Provide default operator
            if (!defined $new->operator) {
                $new->operator('>=');
            }
            # Provide default padding
            if (!defined $new->operatorLeftPad) {
                $new->operatorLeftPad($self->operatorLeftPad // ' ');
            }
            if (!defined $new->versionLeftPad) {
                $new->versionLeftPad($self->versionLeftPad // ' ');
            }
        }
    } else {
        if ($self->type eq 'dependency') {
            die "Inserting a " . $new->type .
                " node between dependencies violates syntax";
        }
        if ($new->isTag and
                (!defined $new->leftpad or !defined $new->seperator)) {
            $new->copyDefaultSpacing($self->findNearestTag);
        }
    }

    $self->insertSibling($self->getIndex, $new);

    return $self;
}

=head2 insertAfter(NEW_NODE)

Insert a new L<RPM::SpecEditor::Node> node after this node. If this is
a dependency node, then the new node must also be a dependency. Unspecified
pads will be copied from a sibling or a space and comma-space will be used.

=cut

sub insertAfter {
    my ($self, $new) = @_;

    if ($new->type eq 'dependency') {
        if ($self->type ne 'dependency') {
            die "Inserting a dependency node next to " . $new->type .
                " node violates syntax. Insert a tag or a Requires-like " .
                "node first.";
        }
        # Provide default rightPad for current node
        if (!defined $self->rightPad) {
            if (my $previous = $self->previousSibling) {
                $self->rightPad($previous->rightPad)
            } else {
                $self->rightPad(', ');
            }
        }
        # Provide default rightPad for new node
        if (!defined $new->rightPad and $self->nextSibling) {
            $new->rightPad($self->rightPad);
        }
        if (defined $new->version) {
            # Provide default operator
            if (!defined $new->operator) {
                $new->operator('>=');
            }
            # Provide default padding
            if (!defined $new->operatorLeftPad) {
                $new->operatorLeftPad($self->operatorLeftPad // ' ');
            }
            if (!defined $new->versionLeftPad) {
                $new->versionLeftPad($self->versionLeftPad // ' ');
            }
        }
    } else {
        if ($self->type eq 'dependency') {
            die "Inserting a " . $new->type .
                " node between dependencies violates syntax";
        }
        if ($new->isTag and
                (!defined $new->leftpad or !defined $new->seperator)) {
            $new->copyDefaultSpacing($self->findNearestTag);
        }
    }

    $self->insertSibling($self->getIndex + 1, $new);

    return $self;
}

=head2 serialize()

Serialize a node into a string. Return list of lines excluding end-of-line
trailers.

=cut

sub serialize() {
    my $self = shift;
    my $type = $self->type;

    if ($type eq 'line') {
        return $self->value;
    } elsif ($type eq 'comment') {
        return $self->leftpad . '#' . $self->separator . $self->value;
    } elsif ($type eq 'tag') {
        return $self->leftpad . $self->name . ':' . $self->separator .
            $self->value;
    } elsif ($self->isDependencyTag) {
        my $value = '';
        for my $dependency ($self->getAllChildren) {
            $value .= $dependency->serialize;
        }
        return ($self->leftpad // ''). $self->type . ':' .
            ($self->separator // '') . $value;
    } elsif ($type eq 'dependency') {
        return $self->symbol . ($self->operatorLeftPad // '') .
            ($self->operator // '') . ($self->versionLeftPad // '') .
            ($self->version // '') . ($self->rightPad // '');
    } else {
        return $self->value;
    }
}

=head2 isTag()

Return true if the node is of type tag or a dependency tag. Otherwise return
false.

=cut

sub isTag {
    my $self = shift;

    if ($self->type eq 'tag') {
        return 1;
    }
    if ($self->isDependencyTag) {
        return 1;
    }
    return 0;
}

=head2 isDependencyTag()

Return true if the node is one of the dependency tags (Provides, BuildRequires,
BuildConflicts, Requires, Recommends, Suggests, Supplements, Conflicts,
Obsoletes). Otherwise returns false.

=cut

sub isDependencyTag {
    my $self = shift;
    my $type = $self->type;
    my @deptypes = (qw( Provides BuildRequires BuildConflicts
            Requires Recommends Suggests Supplements Conflicts Obsoletes));
    for (@deptypes) {
        if ($type eq $_) {
            return 1;
        }
    }
    if ($type eq 'tag') {
        my $name = $self->name;
        for (@deptypes) {
            if ($name eq $_) {
                return 1;
            }
        }
    }
    return 0;
}

=head2 findDependencies(TYPE, SYMBOL)

Find all dependencies of given type (e.g. Requires, Provides) with given
dependency symbol. It returns list of matching dependency nodes in depth-first
search order in the tree. E.g.

    $spec->dom->findDependencies('Requires', qr/^perl\(/);

return:

    Requires: perl(Errno)
    Requires: perl(Errno) >= 1

but not:

    Requires:     perl
    BuildRequies: perl(Errno)

I<TYPE> and I<SYMBOL> can be a string or a regular expression. I<undef> means
to match everything.

=cut

sub findDependencies () {
    my ($self, $type, $symbol) = @_;
    my @accumulator;

    $self->traverse(sub {
            my $node = shift;

            if ($node->type !~ 'dependency') { return; }

            if (defined $type) {
                my $this_type = $node->getParent->type;
                if (ref($type) eq 'Regexp') {
                    if ($this_type !~ $type) { return; }
                } else {
                    if ($this_type ne $type) { return; }
                }
            }

            if (defined $symbol) {
                my $this_symbol = $node->symbol;
                if (ref($symbol) eq 'Regexp') {
                    if ($this_symbol !~ $symbol) { return; }
                } else {
                    if ($this_symbol ne $symbol) { return; }
                }
            }

            push @accumulator, $node;
            return;
        });

    return @accumulator;
}

=head2 findNearestTag()

Find a nearest node that is a tag (or a dependency tag). It starts on the current node and searches backward and forward. If it finds a tag node, return it. Otherwise it returns undef.

=cut

sub findNearestTag {
    my $previous = my $next = shift;
    do {
        if (defined $previous) {
            if ($previous->isTag) {
                return $previous;
            }
            $previous = $previous->previousSibling;
        }
        if (defined $next) {
            if ($next->isTag) {
                return $next;
            }
            $next = $next->nextSibling;
        }
    } while (defined $previous or defined $next);
    return undef;
}

1;

=head1 AUTHOR

Petr Písař <ppisar@redhat.com>

=head1 COPYING

Copyright (C) 2016  Petr Písař <ppisar@redhat.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
